from gint import *
import assets
import game_time

class Entity:
    def __init__(self):
        self.position = [0, 0]

    def Update(self,game):
        # Update logic for the Entity goes here
        pass
    
    def DrawOutside(self,game):
        pass
    def Draw(self,game):
        # Drawing logic for the Entity goes here
        pass

    def Gizmo(self, game):
        pass

class Collider:
    def __init__(self):
        self.offset = [0, 0]
        self.size = [0, 0]

    def Set(self, x, y, w, h):
        self.offset[0] = x
        self.offset[1] = y
        self.size[0] = w
        self.size[1] = h

    def Gizmo(self, game, x, y):
        x = x + self.offset[0]
        y = y + self.offset[1]
        dline(x, y, x + self.size[0], y, C_RED)
        dline(x, y, x, y + self.size[1], C_RED)
        dline(x + self.size[0], y, x + self.size[0], y + self.size[1], C_RED)
        dline(x, y + self.size[1], x + self.size[0], y + self.size[1], C_RED)

class Character(Entity):
    COLLIDER_IDS = {0, 1, 5, 10, 11, 47, 48, 46, 49}
    COLLIDER_IGNORE = {46,47,48, 49}
    def __init__(self):
        super().__init__()
        self.dir = 1
        self.velocity = [0, 0]
        self.acceleration = [0, 0]
        self.health = 100
        self.attack_power = 10
        self.defense = 5
        self.speed = 1.0
        self.collider = Collider()
        self.grounded = False

    def Gizmo(self, game):
        self.collider.Gizmo(game, int(self.position[0] - game.camera[0]), int(self.position[1] - game.camera[1]))

    def Update(self, game):
        self.UpdatePhysics(game)

    def UpdatePhysics(self, game):
        #add gravity
        self.acceleration[1] += 0.1  # Gravity effect
        # Update the position based on velocity and acceleration
        x = (self.velocity[0] + 0.5 * self.acceleration[0])
        y = (self.velocity[1] + 0.5 * self.acceleration[1])
        self.velocity[0] += self.acceleration[0]
        self.velocity[1] += self.acceleration[1]
        self.WanTo(game, self.position[0] + x, self.position[1] + y)
        # Reset acceleration for the next frame
        self.acceleration[0] = 0
        self.acceleration[1] = 0

    def DetectCollision(self,game,p1,p2):
        x_st = int(p1[0] // game.map.tile_size)
        y_st = int(p1[1] // game.map.tile_size)
        x_end = int(p2[0] // game.map.tile_size)
        y_end = int(p2[1] // game.map.tile_size)
        detected = None
        for i in range(max(0, x_st), min(game.map.width, x_end + 1)):
            for j in range(max(0, y_st), min(game.map.height, y_end + 1)):
                id = game.map.Get(i, j)
                if id in self.COLLIDER_IDS:
                    detected = (i, j, id)
                    if id not in self.COLLIDER_IGNORE:
                        return detected
        return detected
    
    def __tl_br__(self, x, y):
        tl = (x + self.collider.offset[0], y +1 + self.collider.offset[1])
        br = (x + self.collider.offset[0] + self.collider.size[0], y + self.collider.offset[1] + self.collider.size[1])
        return tl, br

    def WanTo(self, game, x, y):
        self.grounded = False
        tl, br = self.__tl_br__(x, y)
        res = self.DetectCollision(game,tl, br)
        if res is not None:
            if not res[2] in self.COLLIDER_IGNORE:
                x = self.position[0]
                tl, br = self.__tl_br__(x, y)
            self.velocity[0] = 0
            bas = self.DetectCollision(game,(tl[0],br[1]),(br[0],br[1]))
            if bas is not None: 
                ajuste = 0
                tempY = y
                stair = False
                if bas[2] == 46 and int(tl[0] // game.map.tile_size) == bas[0]:
                    ajuste = int(tl[0]) % game.map.tile_size
                    stair = True
                elif bas[2] == 49 and int(br[0] // game.map.tile_size) == bas[0]:
                    ajuste = game.map.tile_size - (int(br[0]) % game.map.tile_size)
                    stair = True
                if ajuste <= (self.speed * game_time.SpeedCompensation())+0.1:
                    ajuste = 0
                y = (bas[1] * game.map.tile_size) - self.collider.offset[1] - self.collider.size[1] - 1 + ajuste
                if stair and tempY < y:
                    y = tempY
                else:
                    self.grounded = True
                    self.velocity[1] = 0
                    tl, br = self.__tl_br__(x, y)
            haut = self.DetectCollision(game,(tl[0],tl[1]),(br[0],tl[1]))
            if haut is not None:
                if haut[2] == 47:
                    if int(tl[0]) % game.map.tile_size > int(tl[1]) % game.map.tile_size:
                        y = self.position[1] + 1
                        x = self.position[0]
                        self.velocity[1] = 0
                elif haut[2] == 48:
                    if game.map.tile_size - int(tl[0]) % game.map.tile_size > int(tl[1]) % game.map.tile_size:
                        y = self.position[1] + 1
                        x = self.position[0]
                        self.velocity[1] = 0
                else:
                    y = self.position[1] + 1
                    self.velocity[1] = 0
                tl, br = self.__tl_br__(x, y)
        self.position[0] = x
        self.position[1] = y

#IDLE = 0
#RUN = 1
#JUMP = 2
#FALL = 3
#ATTACK = 4
class Player(Character):
    def __init__(self):
        super().__init__()
        self.animation = 0
        self.states_anim = [[0,2, 0.1],[3,8, 0.2],[9,14, 0.2],[14,15, 0.2],[0,4, 0.4]]
        self.state = 0
        self.change_state(0)
        self.speed = 3.0
        self.collider.Set(6, 0+4, 16, 32-4)
        self.smoothX = 0

        self.can_attack = []
        self.attack_power = 40

    def change_state(self, state):
        if state != self.state:
            self.state = state
            self.animation = self.states_anim[state][0]

    def Attack(self, game):
        for entity in self.can_attack:
            if self.position[0] < entity.position[0] and self.dir == 1:
                entity.TakeDamage(self.attack_power)
            elif self.position[0] > entity.position[0] and self.dir == -1:
                entity.TakeDamage(self.attack_power)

    def Update(self,game):
        super().Update(game)
        speed = self.speed * game_time.SpeedCompensation()
        if self.velocity[1] > 1:
            self.change_state(3)
        if self.state <= 1:
            if keypressed(KEY_SHIFT):
                self.change_state(4)
                self.Attack(game)
            elif keydown(KEY_ALPHA):
                tl, br = self.__tl_br__(self.position[0], self.position[1]-1)
                haut = self.DetectCollision(game,(tl[0],tl[1]-4),(br[0],tl[1]-4))
                if haut is None or haut[2] == 47 or haut[2] == 48:
                    self.acceleration[1] = -self.speed*1.25
                    if keydown(KEY_LEFT):
                        self.velocity[0] = -speed
                    elif keydown(KEY_RIGHT):
                        self.velocity[0] = speed
                    self.velocity[0] /= 1.75
                    self.change_state(2)
            elif keydown(KEY_LEFT):
                if self.smoothX > 0:
                    self.smoothX = 0
                self.dir = -1
                self.change_state(1)
                if abs(self.smoothX) < speed:
                    self.smoothX += -speed/8
                self.velocity[0] = self.smoothX
            elif keydown(KEY_RIGHT):
                if self.smoothX < 0:
                    self.smoothX = 0
                self.dir = 1
                self.change_state(1)
                if abs(self.smoothX) < speed:
                    self.smoothX += speed/8
                self.velocity[0] = self.smoothX
            else:
                self.change_state(0)
        elif self.state == 2 or self.state == 3:
            if self.grounded:
                self.change_state(0)
            if self.state == 2 and int(self.animation) == self.states_anim[self.state][1]:
                self.change_state(3)
            if keydown(KEY_LEFT):
                self.dir = -1
                if abs(self.velocity[0]) < 0.2:
                    self.velocity[0] = -0.2
                elif self.velocity[0] > 0:
                    self.velocity[0] /= 1.05
            elif keydown(KEY_RIGHT):
                self.dir = 1
                if abs(self.velocity[0]) < 0.2:
                    self.velocity[0] = 0.2
                elif self.velocity[0] < 0:
                    self.velocity[0] /= 1.05
        elif self.state == 4:
            if int(self.animation) == self.states_anim[self.state][1]:
                self.change_state(0)
        if self.state != 1:
            self.smoothX = 0

    def Draw(self,game):
        self.animation += min(1,self.states_anim[self.state][2]*(game_time.SpeedCompensation()*game_time.RepeatLoop()))
        index = int(self.animation)
        if index > self.states_anim[self.state][1]:
            self.animation = self.states_anim[self.state][0]
            index = self.states_anim[self.state][0]
        if self.state == 4:
            y = 0 if self.dir == 1 else 50
            x = 2 if self.dir == 1 else -18
            game.DrawSubImage(int(self.position[0]+x - game.camera[0]),int(self.position[1]-18 - game.camera[1]), assets.player_attack, 50*index, y, 50, 50)
        else:
            y = 0 if self.dir == 1 else 32
            game.DrawSubImage(int(self.position[0] - game.camera[0]),int(self.position[1] - game.camera[1]), assets.player_basic, 32*index, y, 32, 32)

#WALK = 0
#HIT = 1
#DEATH = 2
#ATTACK = 3
class Skeleton(Character):
    def __init__(self,node):
        super().__init__()
        self.animation = 0
        #[id_start, id_end, speed, x_offset_img, width, tilesheet,(x_offset, y_offset),x_offset_inv]
        self.states_anim = [[0,4, 0.1,0,22,0,(0,0),-5],[0,2, 0.2,110,30,0,(-5,0),-8],[0,2, 0.1,200,33,0,(-8,0),-8],[0,6,0.2,0,43,1,(-5,-4),-21]]
        self.state = 0
        self.position[0]= 20
        self.position[1]= 40
        self.change_state(0)
        self.speed = 0.4
        self.collider.Set(0, 0, 16, 32)
        self.dir = 1

        self.attack_power = 5

        self.target_node = node
        self.can_attack = False
        self.grounded = False
        self.touched = False
        self.cooldown = 0
        self.cooldown_wait = 5

    def change_state(self, state):
        if state != self.state:
            self.cooldown = self.cooldown_wait
            self.state = state
            self.animation = self.states_anim[state][0]

    def TakeDamage(self, damage):
        self.health -= damage
        if self.health <= 0:
            self.change_state(2)
        else:
            self.change_state(1)

    def DetectPlayer(self,game):
        if self.grounded and abs(self.position[0] - 8 - game.player.position[0]) < 32 and abs(self.position[1] - game.player.position[1]) < 16:
            if not self.can_attack:
                self.can_attack = True
                game.player.can_attack.append(self)
            if self.position[0] < game.player.position[0]:
                self.dir = 1
            else:
                self.dir = -1
            self.change_state(3)
        elif self.can_attack:
                self.can_attack = False
                game.player.can_attack.remove(self)
        return self.can_attack
    def Update(self,game):
        if self.state == 1 or self.state == 2:
            return
        if self.health <= 0:
            game.toRemoveEntities.append(self)
            return
        if self.state == 3 and int(self.animation) == 3:
            if not self.touched:
                if self.can_attack:
                    game.ChangeLife(game.player.health-self.attack_power)
                elif self.target_node is None:
                    game.ChangeCrystal(game.Crystal-1)
            self.touched = True
        else:
            self.touched = False
        if self.DetectPlayer(game):
            return
        if self.target_node is None:
            self.change_state(3)
            return
        self.grounded = True
        compensation = game_time.SpeedCompensation()
        pos = [game.graph.nodes[self.target_node][0] * game.map.tile_size, game.graph.nodes[self.target_node][1] * game.map.tile_size-16]
        if abs(self.position[0] - pos[0]) <= 4:
            self.position[0] = pos[0]
            if abs(self.position[1] - pos[1]) <= 6:
                self.position[1] = pos[1]
                self.target_node = game.graph.Next(self.target_node)
                return
            else:
                if self.position[1] < pos[1]:
                    self.position[1] += 1.5 * compensation
                    self.grounded = False
                else:
                    self.position[1] = pos[1]
        else:
            if abs(self.position[1] - pos[1]) > 4:
                if self.position[1] < pos[1]:
                    self.position[1] += self.speed * compensation
                else:
                    self.position[1] -= self.speed * compensation
            if self.position[0] < pos[0]:
                self.dir = 1
                self.change_state(0)
                self.position[0] += self.speed * compensation
            elif self.position[0] > pos[0]:
                self.dir = -1
                self.change_state(0)
                self.position[0] -= self.speed * compensation

    def Draw(self,game):
        next_anim = True
        if self.state == 3:
            if self.cooldown > 0:
                self.cooldown -= 1
                next_anim = False
        else:
            self.cooldown = self.cooldown_wait
        if next_anim:
            self.animation += min(1,self.states_anim[self.state][2]*(game_time.SpeedCompensation()*game_time.RepeatLoop()))
        index = int(self.animation)
        if index > self.states_anim[self.state][1]:
            self.animation = self.states_anim[self.state][0]
            index = self.states_anim[self.state][0]
            if self.state == 3:
                self.cooldown = self.cooldown_wait
            if self.state == 1 or self.state == 2:
                self.change_state(0)
                return
        img = assets.skeleton_basic if self.states_anim[self.state][5] == 0 else assets.skeleton_attack
        y = 0 if self.dir == 1 else img.height // 2
        xo,yo = self.states_anim[self.state][6]
        if self.dir == -1:
            xo = self.states_anim[self.state][7]
        game.DrawSubImage(int(self.position[0] + xo - game.camera[0]),int(self.position[1] + yo- game.camera[1]),
                          img,
                          self.states_anim[self.state][3] + self.states_anim[self.state][4]*index, y,
                          self.states_anim[self.state][4], int(img.height // 2))
        
    def DrawOutside(self,game):
        next_anim = True
        if self.state == 3:
            if self.cooldown > 0:
                self.cooldown -= 1
                next_anim = False
        else:
            self.cooldown = self.cooldown_wait
        if next_anim:
            self.animation += min(1,self.states_anim[self.state][2]*(game_time.SpeedCompensation()*game_time.RepeatLoop()))
        index = int(self.animation)
        if index > self.states_anim[self.state][1]:
            self.animation = self.states_anim[self.state][0]
            index = self.states_anim[self.state][0]
            if self.state == 3:
                self.cooldown = self.cooldown_wait
            if self.state == 1 or self.state == 2:
                self.change_state(0)
class Crystal(Entity):
    def __init__(self):
        super().__init__()
        self.anim = 0

    def Draw(self,game):
        self.anim += min(1, 0.1*(game_time.SpeedCompensation()*game_time.RepeatLoop()))
        if self.anim > 12:
            self.anim = 0
        temp = self.anim if self.anim < 6 else 12 - self.anim
        game.DrawSubImage(int(self.position[0] - game.camera[0]),int(self.position[1] - game.camera[1] - temp), assets.crystal,
                          0, 0, assets.crystal.width, assets.crystal.height)