from gint import *
import assets
from entities import *
import game_time
import random

def DrawMainLogo(x,y):
    x,y = int(x), int(y)
    dsubimage(x, y, assets.ui, 0, 0, 122, 59)
    dsubimage(x+122, y+37, assets.ui, 122, 37, 158-122, 59-37)

def DrawGaugeLife(x,y,percent):
    x,y = int(x), int(y)
    drect(x+15, y+7, x+60, y+9, C_BLACK)
    ext = int((60-15) * percent)
    drect(x+15, y+7, x + 15 + ext, y+9, C_RED)
    dsubimage(x, y, assets.ui, 0, 60, 64, 15)

def DrawGaugeCrystal(x,y,percent):
    x,y = int(x), int(y)
    drect(x+3, y+8, x+(129-68-3), y+10, C_BLACK)
    ext = int((129-68-6) * percent)
    drect(x+3, y+8, x + 3 + ext, y+10, C_RED)
    dsubimage(x, y, assets.ui, 68, 60, 129-68, 15)

def DrawPanel(tl, br):
    dsubimage(tl[0], tl[1], assets.ui, 124, 0, 17, 17)
    dsubimage(tl[0], br[1]-17, assets.ui, 124, 19, 17, 17)
    dsubimage(br[0]-17, tl[1], assets.ui, 124 + 19, 0, 17, 17)
    dsubimage(br[0]-17, br[1]-17, assets.ui, 124 + 19, 19, 17, 17)
    for i in range(tl[0] + 17, br[0] - 16):
        dsubimage(i, tl[1], assets.ui, 124 + 19, 0, 1, 17)
        dsubimage(i, br[1]-17, assets.ui, 124 + 19, 19, 1, 17)
    for i in range(tl[1] + 17, br[1] - 16):
        dsubimage(tl[0], i, assets.ui, 124, 19, 17, 1)
        dsubimage(br[0]-17, i, assets.ui, 124 + 19, 19, 17, 1)
    drect(tl[0]+17, tl[1]+17, br[0]-17, br[1]-17, 43463)

class Map:
    def __init__(self, width=25, height=14, data=None):
        self.width = width
        self.height = height
        if data is not None:
            self.tiles = data
        else:
            self.tiles = [0] * (self.width * self.height)
        self.tile_size = 16
        self.edited = set()
    
    def Load(self, data):
        del self.tiles
        self.tiles = data

    def Set(self, x, y, id):
        self.tiles[y * self.width + x] = id
        self.AddRefresh(x, y)

    def Get(self, x, y):
        return self.tiles[y * self.width + x]
    
    def DrawAll(self):
        for y in range(max(0, Game.Instance.camera[1] // self.tile_size), min(self.height, ((Game.Instance.camera[1] + DHEIGHT) // self.tile_size)), 1):
            y_pos = int((y * self.tile_size) - Game.Instance.camera[1])
            for x in range(max(0, Game.Instance.camera[0] // self.tile_size), min(self.width, ((Game.Instance.camera[0] + DWIDTH) // self.tile_size)+1), 1):
                tile_id = self.tiles[y * self.width + x]
                tile_x = (tile_id % 5) * self.tile_size
                tile_y = (tile_id // 5) * self.tile_size
                dsubimage(int((x * self.tile_size) - Game.Instance.camera[0]), y_pos, assets.tiles, tile_x, tile_y, self.tile_size, self.tile_size)
        Game.Instance.refreshCrystal = True
        Game.Instance.refreshLife = True
    def DrawEdited(self):
        for x,y in self.edited:
            y_pos = int((y * self.tile_size) - Game.Instance.camera[1])
            if y_pos <= 16:
                Game.Instance.refreshCrystal = True
                Game.Instance.refreshLife = True
            tile_id = self.tiles[y * self.width + x]
            tile_x = (tile_id % 5) * self.tile_size
            tile_y = (tile_id // 5) * self.tile_size
            dsubimage(int((x * self.tile_size) - Game.Instance.camera[0]), y_pos, assets.tiles, tile_x, tile_y, self.tile_size, self.tile_size)
        self.edited.clear()

    def AddRefresh(self, x, y):
        if 0 <= x < self.width and 0 <= y < self.height:
            self.edited.add((x, y))

class Graph:
    def __init__(self, nodes=None, edges=None):
        if nodes is not None:
            self.nodes = nodes
        else:
            self.nodes = []
        if edges is not None:
            self.edges = edges
        else:
            self.edges = []

    def Next(self, node):
        if node < len(self.nodes):
            if len(self.edges[node]) == 0:
                return None
            if len(self.edges[node]) == 1:
                return self.edges[node][0]
            if len(self.edges[node]) > 1:
                return random.choice(self.edges[node])
        return None

class Game:
    Instance = None
    def __init__(self,map,graph, spawnable_ids):
        self.__class__.Instance = self
        self.quit = False
        self.map = map
        self.graph = graph
        self.spawnable_ids = spawnable_ids
        self.camera = [0, 0]
        self.player = Player()
        self.entities = [self.player]
        self.toRemoveEntities = []
        self.toDraw = []
        self.Crystal = 100
        self.refreshLife = True
        self.refreshCrystal = True
        self.map.DrawAll()

    def ChangeLife(self, value):
        if value != self.player.health:
            self.refreshLife = True
            self.player.health = value

    def ChangeCrystal(self, value):
        if value != self.Crystal:
            self.refreshCrystal = True
            self.Crystal = value

    def MoveCamera(self, x, y):
        self.camera[0] = x
        self.camera[1] = y
        self.Refresh()

    def Refresh(self):
        self.map.DrawAll()
        self.refreshLife = True
        self.refreshCrystal = True
        game_time.ResetTime()

    def SpawnSkeleton(self):
        node = random.choice(self.spawnable_ids)
        skeleton = Skeleton(node)
        pos = self.graph.nodes[node]
        skeleton.position = [pos[0] * self.map.tile_size, pos[1] * self.map.tile_size-16]
        self.entities.append(skeleton)
        return skeleton

    def Update(self):
        game_time.UpdateTime()
        #compensation du ralentissement
        for i in range(game_time.RepeatLoop()):
            for entity in self.entities:
                entity.Update(self)
            for entity in self.toRemoveEntities:
                if entity in self.entities:
                    self.entities.remove(entity)
                    self.SpawnSkeleton()
            self.toRemoveEntities.clear()
        if (self.player.position[0] - self.camera[0]) < -18:
            self.MoveCamera(self.camera[0] - 25 * self.map.tile_size, self.camera[1])
        if (self.player.position[0] - self.camera[0]) > DWIDTH-14:
            self.MoveCamera(self.camera[0] + 25 * self.map.tile_size, self.camera[1])
        if (self.player.position[1] - self.camera[1]) < -18:
            self.MoveCamera(self.camera[0], self.camera[1] - 14 * self.map.tile_size)
        if (self.player.position[1] - self.camera[1]) > DHEIGHT-14:
            self.MoveCamera(self.camera[0], self.camera[1] + 14 * self.map.tile_size)

    def Draw(self):
        self.map.DrawEdited()
        for entity in self.entities:
            if entity.position[0] - self.camera[0] > -18 \
                and entity.position[0] - self.camera[0] < DWIDTH-14 \
                    and entity.position[1] - self.camera[1] > -18 \
                        and entity.position[1] - self.camera[1] < DHEIGHT-14:
                entity.Draw(self)
            else:
                entity.DrawOutside(self)
        for x, y, image, sx, sy, sw, sh in self.toDraw:
            dsubimage(x, y, image, sx, sy, sw, sh)
        self.toDraw.clear()
        #for entity in self.entities:
        #    entity.Gizmo(self)
        if self.refreshLife:
            self.refreshLife = False
            DrawGaugeLife(0, 0, self.player.health / 100)
        if self.refreshCrystal:
            self.refreshCrystal = False
            DrawGaugeCrystal((DWIDTH/2)-31,0, self.Crystal / 100)

    def UpdateInput(self):
        cleareventflips()
        clearevents()

    def DrawSubImage(self, x, y, image, sx, sy, sw, sh, Refresh=True):
        self.toDraw.append((x, y, image, sx, sy, sw, sh))
        if not Refresh:return
        if y <= 16:
            self.refreshCrystal = True
            self.refreshLife = True
        x += self.camera[0]
        y += self.camera[1]
        x_tile = int(x // self.map.tile_size)
        y_tile = int(y // self.map.tile_size)
        x_end = int((x + sw) // self.map.tile_size)
        y_end = int((y + sh) // self.map.tile_size)
        for i in range(x_tile, x_end + 1):
            for j in range(y_tile, y_end + 1):
                self.map.AddRefresh(i, j)